################################################################################ 
# File: mainframe_dqm.py
# Function: main GUI module using Tkinter
# Author: SK Park
# Revision History
# V0.99a:SK Park    06/28/2019
#    Module Created
# V0.99b: SK Park   07/22/2019
#    Add class filter and -distance option 
################################################################################
import os
import sys
import subprocess
import multiprocessing
#import tkFileDialog
from tkinter import filedialog as tkFileDialog
#import tkMessageBox
from tkinter import messagebox as tkMessageBox
import tkinter as tk     
import tkinter.ttk as ttk  

# Neighborhood distance (m)
MIN_NEIGHBOR_DISATNCE = 0.1
DEFAULT_NEIGHBOR_DISATNCE = 0.5
MAX_NEIGHBOR_DISATNCE = 2

# Neighborhood points
MIN_NEIGHBOR_POINTS = 5
DEFAULT_NEIGHBOR_POINTS = 25
MAX_NEIGHBOR_POINTS = 100

# Number of comparison points
MIN_COMPARISON = 500
DEFAULT_COMPARISON = 1000
MAX_COMPARISON = 100000

# Horrizontal error threshold (m)
MIN_HORIZONTAL_ERROR = 0.01
DEFAULT_HORIZONTAL_ERROR = 0.03
MAX_HORIZONTAL_ERROR = 1.00   

# Vertical error threshold (m)
MIN_VERTICAL_ERROR = 0.01
DEFAULT_VERTICAL_ERROR = 0.04 
MAX_VERTICAL_ERROR = 1.00    
   
MIN_DISCREPANCY_ANGLE=0.01
MAX_DISCREPANCY_ANGLE=1.00
DEFAULT_DISCREPANCY_ANGLE=0.05
    
class MainFrame(tk.Frame):
    
    def __init__(self, parent):
        tk.Frame.__init__(self, parent)
        
        # add menu 
        self.menuBar = tk.Menu(parent)        
        self.helpMenu = tk.Menu(self.menuBar, tearoff=0)
        self.menuBar.add_cascade(label = 'Help', menu=self.helpMenu)
        self.helpMenu.add_command(label='Contents', 
                        command=self.ShowContent)        
        self.helpMenu.add_command(label='About', 
                        command=self.ShowAbout)        
        parent.config(menu=self.menuBar)
        
        for i in range(3):
            parent.rowconfigure(i, weight=1)
            parent.columnconfigure(i, weight=1)                
                
        Frame1 = tk.LabelFrame(parent, text='Select Directory For')
        Frame1.grid(row = 0, column = 0, sticky = 'WENS', padx=10, pady=10) 
        Frame2 = tk.LabelFrame(parent, text='Parameters')
        Frame2.grid(row = 1, column = 0, sticky = 'WENS', padx=10, pady=10)
        Frame3 = tk.Frame(parent)
        Frame3.grid(row = 2, column = 0, sticky = 'WENS', padx=10, pady=10)
        
        # Frame 1
        panel1 = tk.PanedWindow(Frame1)
                
        ### remove if swath mode is not supported
        _ = tk.Label(panel1, text='Data Type:', anchor=tk.NW)
        _.grid(row=0, column=0, sticky = 'WENS', padx=5)
        self.dataTypeVar = tk.IntVar()
        self.swathMode = tk.Radiobutton(panel1, text='Swath', variable=self.dataTypeVar, value=1)
        self.swathMode.grid(row=0, column=1, sticky = 'W')
        self.tileMode = tk.Radiobutton(panel1, text='Tile', variable=self.dataTypeVar, value=2)
        self.tileMode.grid(row=0, column=2, sticky = 'W')
        self.dataTypeVar.set(2)
        ###
        
        buttonData = tk.Button(panel1, text='Data', command=self.SelectInputPath)
        buttonData.grid(row=1, column=0, sticky = 'WENS', padx=10, pady=2)
    
        scrollbar1 = tk.Scrollbar(panel1, orient=tk.HORIZONTAL)
        self.labelData = tk.Entry(panel1, width=60, state=tk.DISABLED, xscrollcommand=scrollbar1.set)
        self.labelData.grid(row=1, column=1, columnspan = 4, sticky = 'WENS', padx=5)
        scrollbar1.config(command=self.labelData.xview)
        scrollbar1.grid(row=2, column=1, columnspan = 4, sticky = 'WENS', padx=5) 
                
        buttonOutput = tk.Button(panel1, text='Output', command=self.SelectOutputPath)
        buttonOutput.grid(row=3, column=0, sticky = 'WENS', padx=10, pady=2)
        scrollbar2 = tk.Scrollbar(panel1, orient=tk.HORIZONTAL)
        self.labelOutput = tk.Entry(panel1, width=60, state=tk.DISABLED, xscrollcommand=scrollbar2.set)
        self.labelOutput.grid(row=3, column=1, columnspan = 4, sticky = 'WENS', padx=5) 
        scrollbar2.config(command=self.labelOutput.xview)
        scrollbar2.grid(row=4, column=1, columnspan = 4, sticky = 'WENS', padx=5)        
        panel1.pack()
        
        # Frame 2
        nb = ttk.Notebook(Frame2)        
        panel2 = tk.PanedWindow(nb)
        nb.add(panel2, text='Configuration')
             
        _ = tk.Label(panel2, text='Unit of measurement:', anchor=tk.NW)
        _.grid(row=0, column=0, sticky = 'W', padx=5, pady=2)
        self.unitVar = tk.IntVar()
        self.unitFoot = tk.Radiobutton(panel2, text='Feet', variable=self.unitVar, value=1)
        self.unitFoot.grid(row=0, column=1, sticky = 'W')
        self.unitMeter = tk.Radiobutton(panel2, text='Meter', variable=self.unitVar, value=2)
        self.unitMeter.grid(row=0, column=2, sticky = 'W')
        self.unitVar.set(2)
                                 
        _ = tk.Label(panel2, text='Neighborhood parameter:', anchor=tk.NW)
        _.grid(row=1, column=0, sticky = 'W', padx=5, pady=2)
        self.neighboringVar = tk.IntVar() 
        self.unitDistance = tk.Radiobutton(panel2, text='Distance(m)', variable=self.neighboringVar, value=1)
        self.unitDistance.grid(row=1, column=1, sticky = 'W')
        self.NDistanceVal = tk.DoubleVar()
        NDistance = tk.Entry(panel2, width=5, textvariable=self.NDistanceVal)  
        self.NDistanceVal.set(DEFAULT_NEIGHBOR_DISATNCE)
        NDistance.grid(row=1, column=2, sticky = 'W', padx=5)
        self.unitPoints = tk.Radiobutton(panel2, text='Points', variable=self.neighboringVar, value=2)
        self.unitPoints.grid(row=1, column=3, sticky = 'W')
        self.NPointsVal = tk.IntVar()
        NPointsEntry = tk.Entry(panel2, width=5, textvariable=self.NPointsVal)  
        self.NPointsVal.set(DEFAULT_NEIGHBOR_POINTS)
        NPointsEntry.grid(row=1, column=4, sticky = 'W', padx=5)      
        self.neighboringVar.set(2)
  
        _ = tk.Label(panel2, text='Number of CPU for processing:', anchor=tk.NW)
        _.grid(row=2, column=0, sticky = 'W', padx=5, pady=2)
        self.Ncores = tk.IntVar()
        n_cores = multiprocessing.cpu_count()  
        if(n_cores>10):
            n_cores=10       
        self.Ncores = tk.Spinbox(panel2, width=5, from_=1, to=n_cores)
        self.Ncores.delete(0, 'end')
        if n_cores > 3:
            self.Ncores.insert(0, 3)   
        self.Ncores.grid(row=2, column=1, sticky = 'WENS', padx=5, pady=2)
                
        _ = tk.Label(panel2, text='Number of comparison points:', anchor=tk.NW)
        _.grid(row=3, column=0, sticky = 'WENS', padx=5, pady=2)
        self.NComparisonVal = tk.IntVar()
        NComparisonEntry = tk.Entry(panel2, width=10, textvariable=self.NComparisonVal)  
        self.NComparisonVal.set(DEFAULT_COMPARISON)
        NComparisonEntry.grid(row=3, column=1, sticky = 'WENS', padx=5, pady=2)
              
        panel2_2 = tk.PanedWindow(nb)
        nb.add(panel2_2, text='Advanced')   
        
        self.DropClassVal = tk.IntVar()
        FilterCB = tk.Checkbutton(panel2_2, text='Drop Classes', variable=self.DropClassVal, command=self.setDropClass)
        FilterCB.grid(row=0, column=0, sticky = 'W', padx=5, pady=2)
        self.DropClasses = tk.StringVar()
        self.DropClassEntry = tk.Entry(panel2_2, textvariable=self.DropClasses, state='disabled')        
        self.DropClassEntry.grid(row=0, column=1, columnspan = 2, sticky = 'WENS', padx=5, pady=2)
             
        _ = tk.Label(panel2_2, text='Horizontal error threshold(m):', anchor=tk.NW)
        _.grid(row=1, column=0, sticky = 'WENS', padx=5, pady=2)
        self.HerrorVal = tk.DoubleVar()
        HerrEntry = tk.Entry(panel2_2, textvariable=self.HerrorVal)        
        HerrEntry.grid(row=1, column=1, columnspan = 2, sticky = 'WENS', padx=5, pady=2)
        self.HerrorVal.set(DEFAULT_HORIZONTAL_ERROR)    
       
        _ = tk.Label(panel2_2, text='Vertical error threshold(m):', anchor=tk.NW)        
        _.grid(row=2, column=0, sticky = 'WENS', padx=5, pady=2)     
        self.VerrorVal = tk.DoubleVar()
        VerrEntry = tk.Entry(panel2_2, textvariable=self.VerrorVal)        
        VerrEntry.grid(row=2, column=1, columnspan = 2, sticky = 'WENS', padx=5, pady=2)
        self.VerrorVal.set(DEFAULT_VERTICAL_ERROR)
        
        _ = tk.Label(panel2_2, text='Discrepancy angle:', anchor=tk.NW)        
        _.grid(row=3, column=0, sticky = 'WENS', padx=5, pady=2)     
        self.AngleVal = tk.DoubleVar()
        AngleEntry = tk.Entry(panel2_2, textvariable=self.AngleVal)        
        AngleEntry.grid(row=3, column=1, columnspan = 2, sticky = 'WENS', padx=5, pady=2)
        self.AngleVal.set(DEFAULT_DISCREPANCY_ANGLE)
        _ = tk.Label(panel2_2, text="", anchor=tk.NW)
        _.grid(row=3, column=3, sticky = 'WENS', pady=2) 
        nb.pack()
        
        # Frame 3
        panel3 = tk.PanedWindow(Frame3)
        buttonRun = tk.Button(panel3, text='Run', command=self.RunDQM)
        buttonRun.grid(row=0, column=0, sticky = 'WENS', padx=10, pady=2)
        buttonCancel = tk.Button(panel3, text='Cancel', command=self.CancelDQM)
        buttonCancel.grid(row=0, column=1, sticky = 'WENS', padx=10, pady=2)
        panel3.pack()
        
        self.InputPath = ''
        self.OutputPath = ''
        self.files = []
        self.okPressed = 0
        self.drop_class = ''
        self.drop_extended_class = ''
    
    def checkDropClass(self):
        self.drop_class = ''
        self.drop_extended_class = ''
        classes = self.DropClasses.get().split()
        for class_code_str in classes:
            try:
                class_code = int(class_code_str)
            except ValueError:
                msg = 'Class code is not a number %s' % (class_code_str)
                tkMessageBox.showerror('Drop Class', msg)
                return 0
            class_code = int(class_code_str)        
            if class_code < 0 or class_code > 255:
                msg = 'Invalid class code %d' % (class_code)
                tkMessageBox.showerror('Drop Class', msg)
                return 0
            if class_code < 32:
                self.drop_class = self.drop_class + class_code_str + ' '
            else:
                self.drop_extended_class = self.drop_extended_class + class_code_str + ' '
        return 1
        
    def setDropClass(self):
        if self.DropClassVal.get() is 1:
            self.DropClasses.set("7 18 ")
            self.DropClassEntry.config(state='normal')
        else:
            self.drop_class = ''
            self.drop_extended_class = ''
            self.DropClasses.set("")
            self.DropClassEntry.config(state='disabled')              
                       
    def SelectInputPath(self): 
        self.InputPath = tkFileDialog.askdirectory(initialdir="/",title='Choose Directory of LAS Files')
        if len(self.InputPath) > 0:
            self.checkDataFileList()
            if len(self.files) > 0:
                self.labelData.configure(state=tk.NORMAL)
                self.labelData.insert(0, self.InputPath)
                self.labelData.configure(state=tk.DISABLED)
            
    def SelectOutputPath(self):
        self.OutputPath = tkFileDialog.askdirectory(initialdir="/",title='Choose Directory for Output')
        if len(self.OutputPath) > 0:
            self.labelOutput.configure(state=tk.NORMAL)
            self.labelOutput.insert(0, self.OutputPath)
            self.labelOutput.configure(state=tk.DISABLED)
            
    def ShowAbout(self):
        tkMessageBox.showinfo('About', 'EROS Lidar DQM V2.00')
        
    def ShowContent(self):
        doc = "DQM_Operational_Testing 2018-2-6.pptx"
        if os.path.exists(doc):
            subprocess.Popen(doc, shell=True)
        
    def RunDQM(self):
        # check any input error
        self.okPressed = 0
        if len(self.InputPath) == 0:
            tkMessageBox.showerror('Input Data Path', 'Select Input Data Path!')
            return
        if len(self.OutputPath) == 0:
            tkMessageBox.showerror('Output Path', 'Select Output Path!')
            return
        
        param = self.neighboringVar.get()
        if param is 2:
            try:
                val = self.getNeighborPoints()
            except ValueError:
                tkMessageBox.showerror('Error', 'Number of neighboring points: Invalid number')
                return
            if (val < MIN_NEIGHBOR_POINTS) or (val > MAX_NEIGHBOR_POINTS):            
                msg = 'Enter between %d and %d' % (MIN_NEIGHBOR_POINTS, MAX_NEIGHBOR_POINTS)
                tkMessageBox.showerror('Number of neighboring points', msg)
                return
        else:
            try:
                val = self.getNeighborDistance()
            except ValueError:
                tkMessageBox.showerror('Error', 'Neighboring distance: Invalid number')
                return
            if (val < MIN_NEIGHBOR_DISATNCE) or (val > MAX_NEIGHBOR_DISATNCE):            
                msg = 'Enter between %d and %d' % (MIN_NEIGHBOR_DISATNCE, MAX_NEIGHBOR_DISATNCE)
                tkMessageBox.showerror('Neighborhood distance', msg)
                return
        
        try:
            val = self.NComparisonVal.get()
        except ValueError:
            tkMessageBox.showerror('Error', 'Number of comparison points: Invalid number')
            return            
        if (val < MIN_COMPARISON) or (val > MAX_COMPARISON):            
            msg = 'Enter between %d and %d' % (MIN_COMPARISON, MAX_COMPARISON)
            tkMessageBox.showerror('Number of comparison points', msg)
            return
        
        val = self.DropClassVal.get()
        if val is 1:
            ok = self.checkDropClass()
            if ok is 0:
                return   
        
        try:
            val = self.HerrorVal.get()
        except ValueError:
            tkMessageBox.showerror('Error', 'Horizontal error threshold: Invalid number')
            return
        if (val < MIN_HORIZONTAL_ERROR) or (val > MAX_HORIZONTAL_ERROR):            
            msg = 'Enter between %d and %d' % (MIN_HORIZONTAL_ERROR, MAX_HORIZONTAL_ERROR)
            tkMessageBox.showerror('Horizontal error threshold', msg)
            return
        
        try:
            val = self.VerrorVal.get()
        except ValueError:
            tkMessageBox.showerror('Error', 'Vertical error threshold: Invalid number')
            return
        if (val < MIN_VERTICAL_ERROR) or (val > MAX_VERTICAL_ERROR):            
            msg = 'Enter between %d and %d' % (MIN_VERTICAL_ERROR, MAX_VERTICAL_ERROR)
            tkMessageBox.showerror('Vertical error threshold', msg)
            return        
        
        try:
            val = self.AngleVal.get()
        except ValueError:
            tkMessageBox.showerror('Error', 'Discrepancy angle: Invalid number')
            return
        if (val < MIN_DISCREPANCY_ANGLE) or (val > MAX_DISCREPANCY_ANGLE):            
            msg = 'Enter between %d and %d' % (MIN_DISCREPANCY_ANGLE, MAX_DISCREPANCY_ANGLE)
            tkMessageBox.showerror('Discrepancy angle', msg)
            return
        
        self.okPressed = 1
        self.quit()     
    
    def CancelDQM(self):
        self.okPressed = 0
        self.quit()
        
    def getInputPath(self):
        return self.InputPath
    
    def checkDataFileList(self):
        self.files=[]
        for path, _, files in os.walk(self.InputPath):
            for f in files:
                if f.endswith('.laz') or f.endswith('.las') or f.endswith('.LAS'):
                    self.files.append(os.path.join(path, f))
        
        if len(self.files) == 0:
            tkMessageBox.showwarning('Select Input Data', 'No LAS files are found')
    
    def getDataFileList(self):        
        return self.files
    
    def getOutputPath(self):
        return self.OutputPath
    
    def getDataType(self):
        return self.dataTypeVar.get()
    
    def getUnitOfMeasurement(self):
        return self.unitVar.get()
    
    def getNumCPUcores(self):
        return self.Ncores.get()
    
    def getNeighboringParam(self):
        return self.neighboringVar.get()
    
    def getNeighborPoints(self):
        return self.NPointsVal.get()
    
    def getNeighborDistance(self):
        return self.NDistanceVal.get()
    
    def getNumComparison(self):
        return self.NComparisonVal.get()
    
    def getHErrorThreshold(self):
        return self.HerrorVal.get()
    
    def getVErrorThreshold(self):
        return self.VerrorVal.get()      
    
    def getDiscrepancyAngle(self):
        return self.AngleVal.get()    
    
    def getDropCLass(self):
        return self.DropClassVal
    
    def getDropClassCode(self):
        return self.drop_class
    
    def getDropExtendedClassCode(self):
        return self.drop_extended_class  
    
if __name__ == '__main__':
    # create aTk root window
    top = tk.Tk()
    top.title("Eros Lidar DQM")
    
    try:
        img = tk.PhotoImage(file='image\\Usgs_logo.png')
        top.tk.call('wm', 'iconphoto', top._w, img)
    except:
        print ('EROS logo icon not found!')
    
    # build and realize main frame
    app = MainFrame(top)
    
    # enter eventloop
    app.mainloop()
    
    if app.okPressed is 0:
        print ('DQM is canceled')
        sys.exit(0)
        
    print ('Input Path = %s' % app.getInputPath())
    print ('Output Path = %s' % app.getOutputPath())
    print ('Data Type = %d' % app.getDataType())
    print ('Unit of Measurement = %d' % app.getUnitOfMeasurement())
    print ('Num CPU cores = %d' % int(app.getNumCPUcores()))
    print ('Num neighborhood points = %d' % int(app.getNeighborPoints()))
    print ('Num neighborhood distance = %s' % app.getNeighborDistance())
    print ('Num comparison points = %d' % int(app.getNumComparison()))
    print ('Horrizontal error = %s' % app.getHErrorThreshold())
    print ('Vertical error = %s' % app.getVErrorThreshold())
    print ('Discrepancy Angle = %s' % app.getDiscrepancyAngle())
    print ('Drop Class = %s' % app.getDropClassCode())
    print ('Drop Extended Class = %s' % app.getDropExtendedClassCode())
  
        
